
"""
This provides the Theano Op `SprintErrorSigOp` to get a loss and error signal
which is calculated via Sprint.
And there are helper classes to communicate with the Sprint subprocess
to transfer the posteriors and get back the loss and error signal.
It uses the SprintControl Sprint interface for the communication.
"""

from __future__ import print_function

import numpy
import sys
import os
import atexit
import signal
import typing
from threading import RLock, Thread
import returnn.util.task_system as task_system
from returnn.util.task_system import Pickler, Unpickler, numpy_set_unused
from returnn.util.basic import eval_shell_str, make_hashable, close_all_fds_except
from returnn.log import log


class SprintSubprocessInstance:
  """
  The Sprint instance which is used to calculate the error signal.
  Communication is over a pipe. We pass the fds via cmd-line to the child proc.
  Basic protocol with the subprocess (encoded via pickle):
    P2C: tuple (cmd, *cmd_args). cmd is any str.
    C2P: tuple (status, *res_args). status == "ok" if no error.
  Commands:
    "init", name, version -> "ok", child_name, version
    "exit" -> (exit)
    "get_loss_and_error_signal", seg_name, seg_len, posteriors -> "ok", loss, error_signal
      Numpy arrays encoded via TaskSystem.Pickler (which is optimized for Numpy).
  On the Sprint side, we handle this via the SprintControl Sprint interface.
  """

  Version = 1  # increase when some protocol changes

  # Keep argument names as is, as these are coming directly from a user config file.
  # noinspection PyPep8Naming
  def __init__(self, sprintExecPath, minPythonControlVersion=2, sprintConfigStr="", sprintControlConfig=None,
               usePythonSegmentOrder=True):
    """
    :param str sprintExecPath: this executable will be called for the sub proc.
    :param int minPythonControlVersion: will be checked in the subprocess. via Sprint PythonControl
    :param str sprintConfigStr: passed to Sprint as command line args.
      can have "config:" prefix - in that case, looked up in config.
      handled via eval_shell_str(), can thus have lazy content (if it is callable, will be called).
    :param dict[str]|None sprintControlConfig: passed to SprintControl.init().
    """
    assert os.path.exists(sprintExecPath)
    self.sprintExecPath = sprintExecPath
    self.minPythonControlVersion = minPythonControlVersion
    if sprintConfigStr.startswith("config:"):
      from returnn.config import get_global_config
      config = get_global_config()
      assert config
      sprintConfigStr = config.typed_dict[sprintConfigStr[len("config:"):]]
    self.sprintConfig = eval_shell_str(sprintConfigStr)
    self.sprintControlConfig = sprintControlConfig
    self.usePythonSegmentOrder = usePythonSegmentOrder
    self.child_pid = None  # type: typing.Optional[int]
    self.parent_pid = os.getpid()
    # There is no generic way to see whether Python is exiting.
    # This is our workaround. We check for it in self.run_inner().
    self.python_exit = False
    atexit.register(self.exit_handler)
    self._cur_seg_name = None
    self._cur_posteriors_shape = None
    self.is_calculating = False
    self.init()

  def _exit_child(self, should_interrupt=False):
    if self.child_pid:
      interrupt = False
      expected_exit_status = 0 if not self.python_exit else None
      if self._join_child(wait=False, expected_exit_status=expected_exit_status) is False:  # Not yet terminated.
        interrupt = should_interrupt
        if interrupt:
          print("SprintSubprocessInstance: interrupt child proc %i" % self.child_pid, file=log.v5)
          os.kill(self.child_pid, signal.SIGKILL)
        else:
          # noinspection PyBroadException
          try:
            self._send(("exit",))
          except Exception:
            pass
      else:
        self.child_pid = None
      try:
        self.pipe_p2c[1].close()
      except IOError:
        pass
      try:
        self.pipe_c2p[0].close()
      except IOError:
        pass
      if self.child_pid:
        self._join_child(wait=True, expected_exit_status=0 if not interrupt else None)
        self.child_pid = None

  # noinspection PyMethodMayBeStatic
  def _env_update_child(self):
    theano_flags = {key: value for (key, value)
                    in [s.split("=", 1) for s in os.environ.get("THEANO_FLAGS", "").split(",") if s]}

    # First set some sane default for compile dir.
    theano_flags.setdefault("compiledir_format",
                            "compiledir_%(platform)s-%(processor)s-%(python_version)s-%(python_bitwidth)s")
    compiledir_format = theano_flags["compiledir_format"]
    p = compiledir_format.find("--dev-")  # Device.startProc might have added that.
    if p >= 0:
      compiledir_format = compiledir_format[:p]
    compiledir_format += "--sprint-sub"
    theano_flags["compiledir_format"] = compiledir_format
    theano_flags["device"] = "cpu"  # Force CPU.
    theano_flags["force_device"] = True
    os.environ["THEANO_FLAGS"] = ",".join(["%s=%s" % (key, value) for (key, value) in sorted(theano_flags.items())])

  def _start_child(self):
    assert self.child_pid is None
    self.pipe_c2p = self._pipe_open()
    self.pipe_p2c = self._pipe_open()
    args = self._build_sprint_args()
    print("SprintSubprocessInstance: exec", args, file=log.v5)

    pid = os.fork()
    if pid == 0:  # child
      print("SprintSubprocessInstance: starting, pid %i" % os.getpid(), file=log.v5)
      # noinspection PyBroadException
      try:
        self._env_update_child()
        sys.stdin.close()  # Force no tty stdin.
        self.pipe_c2p[0].close()
        self.pipe_p2c[1].close()
        close_all_fds_except([0, 1, 2, self.pipe_c2p[1].fileno(), self.pipe_p2c[0].fileno()])
        os.execv(args[0], args)  # Does not return if successful.
      except BaseException:
        print("SprintSubprocessInstance: Error when starting Sprint %r." % args, file=log.v1)
        sys.excepthook(*sys.exc_info())
      finally:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        os._exit(1)
        return  # Not reached.

    # parent
    self.pipe_c2p[1].close()
    self.pipe_p2c[0].close()
    self.child_pid = pid

    try:
      self._send(("init", "SprintSubprocessInstance", self.Version))
      ret = self._read()
      assert ret[0] == "ok" and len(ret) >= 3 and ret[2] == self.Version
    except Exception:
      print("SprintSubprocessInstance: Sprint child process (%r) caused an exception." % args, file=log.v1)
      sys.excepthook(*sys.exc_info())
      raise Exception("SprintSubprocessInstance Sprint init failed")

  # noinspection PyMethodMayBeStatic
  def _pipe_open(self):
    readend, writeend = os.pipe()
    if hasattr(os, "set_inheritable"):
      # https://www.python.org/dev/peps/pep-0446/
      os.set_inheritable(readend, True)
      os.set_inheritable(writeend, True)
    readend = os.fdopen(readend, "rb", 0)
    writeend = os.fdopen(writeend, "wb", 0)
    return readend, writeend

  @property
  def _my_python_mod_path(self):
    from returnn import __root_dir__
    return __root_dir__

  def _build_sprint_args(self):
    config_str = "c2p_fd:%i,p2c_fd:%i" % (
        self.pipe_c2p[1].fileno(), self.pipe_p2c[0].fileno())
    config_str += ",minPythonControlVersion:%i" % self.minPythonControlVersion
    if task_system.SharedMemNumpyConfig["enabled"]:
      config_str += ",EnableAutoNumpySharedMemPickling:True"
    if self.sprintControlConfig:
      config_str += "," + ",".join(["%s:%s" % (k, v) for (k, v) in sorted(self.sprintControlConfig.items())])
    my_mod_name = "returnn.sprint.control"
    args = [
      self.sprintExecPath,
      # Enable Sprint PythonControl
      "--*.python-control-enabled=true",
      # Sprint PythonControl or PythonTrainer
      "--*.pymod-path=%s" % self._my_python_mod_path,
      "--*.pymod-name=%s" % my_mod_name,
      "--*.pymod-config=%s" % config_str
    ]
    if self.usePythonSegmentOrder:
      args += [
        # Sprint PythonSegmentOrder
        "--*.python-segment-order=true",
        "--*.python-segment-order-pymod-path=%s" % self._my_python_mod_path,
        "--*.python-segment-order-pymod-name=%s" % my_mod_name,
        "--*.python-segment-order-config=%s" % config_str,
        "--*.python-segment-order-allow-copy=false"
      ]
    args += self.sprintConfig
    return args

  def _send(self, v):
    assert os.getpid() == self.parent_pid
    p = self.pipe_p2c[1]  # see _start_child
    Pickler(p).dump(v)

  def _read(self):
    assert os.getpid() == self.parent_pid
    p = self.pipe_c2p[0]  # see _start_child
    return Unpickler(p).load()

  def _poll(self):
    assert os.getpid() == self.parent_pid
    p = self.pipe_c2p[0]  # see _start_child
    from select import select
    ready, _, _ = select([p.fileno()], [], [], 0)
    return bool(ready)

  def _join_child(self, wait=True, expected_exit_status=None):
    assert self.child_pid
    options = 0 if wait else os.WNOHANG
    pid, exit_status = os.waitpid(self.child_pid, options)
    if not wait and pid == 0:
      return False
    assert pid == self.child_pid
    if expected_exit_status is not None:
      assert exit_status == expected_exit_status, "Sprint exit code is %i" % exit_status
    return True

  def get_loss_and_error_signal__send(self, seg_name, seg_len, log_posteriors):
    """
    :param str seg_name: the segment name (seq_tag)
    :param int seg_len: the segment length in frames
    :param numpy.ndarray log_posteriors: 2d (time,label) float array, log probs
    """
    assert not self.is_calculating
    assert seg_name
    self._cur_seg_name = seg_name
    assert seg_len == log_posteriors.shape[0]
    self._cur_posteriors_shape = log_posteriors.shape
    try:
      self._send(("get_loss_and_error_signal", seg_name, seg_len, log_posteriors.astype("float32", copy=False)))
    except (IOError, EOFError):
      raise
    else:
      self.is_calculating = True

  def get_loss_and_error_signal__have_data(self):
    """
    :rtype: bool
    """
    assert self.is_calculating
    return self._poll()

  def get_loss_and_error_signal__read(self):
    """
    :rtype (str, float, numpy.ndarray)
    :returns (seg_name, loss, error_signal). error_signal has the same shape as posteriors.
    """
    assert self.is_calculating
    try:
      ret = self._read()
    except (IOError, EOFError):
      raise
    else:
      self.is_calculating = False
    assert ret[0] == "ok" and len(ret) == 3, "Got unexpected return: %r" % (ret,)
    loss = ret[1]
    error_signal = ret[2]
    assert error_signal.shape == self._cur_posteriors_shape
    return self._cur_seg_name, loss, error_signal

  def exit_handler(self):
    """
    Called at exit. Exit child.
    """
    assert os.getpid() == self.parent_pid
    self.python_exit = True
    self._exit_child(should_interrupt=True)

  def init(self):
    """
    Init. (Re)Start child.
    """
    self._exit_child()
    self._start_child()


class ReaderThread(Thread):
  """
  Sprint reader thread.
  """
  def __init__(self, instance, instance_idx, batch_idxs, tags, seq_lengths, log_posteriors,
               batch_loss, batch_error_signal):
    """
    :param SprintSubprocessInstance instance:
    :param int instance_idx:
    :param list[int] batch_idxs:
    :param list[str] tags: seq names, length = batch
    :param numpy.ndarray seq_lengths: 1d (batch)
    :param numpy.ndarray log_posteriors: 3d (time,batch,label)
    :param numpy.ndarray batch_loss: 1d (batch). will write result into it.
    :param numpy.ndarray batch_error_signal: 3d (time,batch,label). will write results into it.
    """
    super(ReaderThread, self).__init__(
      name="SprintErrorSignals reader thread for Sprint instance %i" % instance_idx)
    self.daemon = True
    self.instance_idx = instance_idx
    self.instance = instance
    self.batch_idxs = batch_idxs
    self.tags = tags
    self.seq_lengths = seq_lengths
    self.log_posteriors = log_posteriors
    self.batch_loss = batch_loss
    self.batch_error_signal = batch_error_signal
    self.exception = None
    self.start()

  def run(self):
    """
    Main thread func.
    """
    try:
      for b in self.batch_idxs:
        self.instance.get_loss_and_error_signal__send(
          seg_name=self.tags[b], seg_len=self.seq_lengths[b],
          log_posteriors=self.log_posteriors[:self.seq_lengths[b], b])
        seg_name, loss, error_signal = self.instance.get_loss_and_error_signal__read()
        assert seg_name == self.tags[b]
        self.batch_loss[b] = loss
        self.batch_error_signal[:self.seq_lengths[b], b] = error_signal
        numpy_set_unused(error_signal)
    except Exception as exc:
      self.exception = exc


class SprintInstancePool:
  """
  This is a pool of Sprint instances.
  First, for each unique sprint_opts, there is a singleton
    which can be accessed via get_global_instance.
  Then, this can be used in multiple ways.
    (1) get_batch_loss_and_error_signal.
    (2) ...
  """

  class_lock = RLock()
  global_instances = {}  # sprint_opts -> SprintInstancePool instance

  @classmethod
  def get_global_instance(cls, sprint_opts):
    """
    :param dict[str] sprint_opts:
    :rtype: SprintInstancePool
    """
    sprint_opts = make_hashable(sprint_opts)
    with cls.class_lock:
      if sprint_opts in cls.global_instances:
        return cls.global_instances[sprint_opts]
      instance = SprintInstancePool(sprint_opts=sprint_opts)
      cls.global_instances[sprint_opts] = instance
      return instance

  def __init__(self, sprint_opts):
    """
    :param dict[str] sprint_opts:
    """
    # The lock will not be acquired automatically on the public functions here as there is the valid
    # usage that only one thread will access it anyway.
    # So, take care of acquiring this lock yourself whenever you call here potentially from multiple threads.
    # All the code is not thread-safe, so this is important!
    self.lock = RLock()
    assert isinstance(sprint_opts, dict)
    sprint_opts = sprint_opts.copy()
    self.max_num_instances = int(sprint_opts.pop("numInstances", 1))
    self.sprint_opts = sprint_opts
    self.instances = []  # type: typing.List[SprintSubprocessInstance]

  def _maybe_create_new_instance(self):
    """
    :rtype: SprintSubprocessInstance|None
    """
    if len(self.instances) < self.max_num_instances:
      self.instances.append(SprintSubprocessInstance(**self.sprint_opts))
      return self.instances[-1]
    return None

  def _get_instance(self, i):
    """
    :param int i:
    :rtype: SprintSubprocessInstance
    """
    assert i < self.max_num_instances
    if i >= len(self.instances):
      assert i == len(self.instances)
      self._maybe_create_new_instance()
    return self.instances[i]

  def get_batch_loss_and_error_signal(self, log_posteriors, seq_lengths, tags=None):
    """
    :param numpy.ndarray log_posteriors: 3d (time,batch,label)
    :param numpy.ndarray seq_lengths: 1d (batch)
    :param list[str] tags: seq names, length = batch
    :rtype (numpy.ndarray, numpy.ndarray)
    :returns (loss, error_signal). error_signal has the same shape as posteriors.
    loss is a 1d-array (batch).

    Note that this accesses some global references, like global current seg info,
    via the current Device instance.
    Thus this is expected to be run from the Device host proc,
      inside from SprintErrorSigOp.perform.
    This also expects that we don't have chunked seqs.
    """
    assert seq_lengths.ndim == 1
    assert log_posteriors.ndim == 3
    n_batch = seq_lengths.shape[0]
    assert n_batch == log_posteriors.shape[1]

    if tags is None:
      raise NotImplementedError("This feature was removed with dropped Theano support")
    assert len(tags) == n_batch

    batch_loss = numpy.zeros((n_batch,), dtype="float32")
    batch_error_signal = numpy.zeros_like(log_posteriors, dtype="float32")

    # greedy solution to the scheduling problem
    sorted_length = sorted(enumerate(seq_lengths), key=lambda x: x[1], reverse=True)
    jobs = [[] for _ in range(self.max_num_instances)]
    joblen = [0] * self.max_num_instances
    for i, l in sorted_length:
      j = min(enumerate(joblen), key=lambda x: x[1])[0]  # noqa
      jobs[j].append(i)
      joblen[j] += l

    if self.max_num_instances > 1:
      threads = [
        ReaderThread(
          self._get_instance(i), i, jobs[i], tags, seq_lengths, log_posteriors, batch_loss, batch_error_signal)
        for i in range(self.max_num_instances)]
      for i, thread in enumerate(threads):
        thread.join()
        if thread.exception:
          raise thread.exception
    else:
      # Very simple parallelism. We must avoid any form of multi-threading
      # because this can be problematic with Theano.
      # See: https://groups.google.com/forum/#!msg/theano-users/Pu4YKlZKwm4/eNcAegzaNeYJ
      # We also try to keep it simple here.
      for bb in range(0, n_batch, self.max_num_instances):
        for i in range(self.max_num_instances):
          b = bb + i
          if b >= n_batch:
            break
          instance = self._get_instance(i)
          instance.get_loss_and_error_signal__send(
            seg_name=tags[b], seg_len=seq_lengths[b], log_posteriors=log_posteriors[:seq_lengths[b], b])
        for i in range(self.max_num_instances):
          b = bb + i
          if b >= n_batch:
            break
          instance = self._get_instance(i)
          seg_name, loss, error_signal = instance.get_loss_and_error_signal__read()
          assert seg_name == tags[b]
          batch_loss[b] = loss
          batch_error_signal[:seq_lengths[b], b] = error_signal
          numpy_set_unused(error_signal)
    return batch_loss, batch_error_signal

  def get_automata_for_batch(self, tags):
    """
    :param list[str]|numpy.ndarray tags: sequence names, used for Sprint (ndarray of shape (batch, max_str_len))
    :return: (edges, weights, start_end_states). all together in one automaton.
      edges are of shape (4, num_edges), each (from, to, emission-idx, seq-idx), of dtype uint32.
      weights are of shape (num_edges,), of dtype float32.
      start_end_states are of shape (2, batch), each (start,stop) state idx, batch = len(tags), of dtype uint32.
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    """
    all_num_states = [None] * len(tags)  # type: typing.List[typing.Optional[int]]
    all_num_edges = [None] * len(tags)  # type: typing.List[typing.Optional[int]]
    all_edges = [None] * len(tags)  # type: typing.List[typing.Optional[numpy.ndarray]]
    all_weights = [None] * len(tags)  # type: typing.List[typing.Optional[numpy.ndarray]]
    for bb in range(0, len(tags), self.max_num_instances):
      for i in range(self.max_num_instances):
        b = bb + i
        if b >= len(tags):
          break
        instance = self._get_instance(i)
        if isinstance(tags[0], str):
          segment_name = tags[b]
        elif isinstance(tags[0], bytes):
          segment_name = tags[b].decode()
        else:
          segment_name = tags[b].view('S%d' % tags.shape[1])[0]
        assert isinstance(segment_name, str)
        # noinspection PyProtectedMember
        instance._send(("export_allophone_state_fsa_by_segment_name", segment_name))
      for i in range(self.max_num_instances):
        b = bb + i
        if b >= len(tags):
          break
        instance = self._get_instance(i)
        # noinspection PyProtectedMember
        r = instance._read()
        if r[0] != 'ok':
          raise RuntimeError(r[1])
        num_states, num_edges, edges, weights = r[1:]
        all_num_states[b] = num_states
        all_num_edges[b] = num_edges
        all_edges[b] = edges.reshape((3, num_edges))  # (from, to, emission-idx) for each edge, uint32
        all_weights[b] = weights  # for each edge, float32
    state_offset = 0
    for idx in range(len(all_edges)):
      num_edges = all_num_edges[idx]
      all_edges[idx][0:2, :] += state_offset
      state_offset += all_num_states[idx]
      # add sequence_idx. becomes (from, to, emission-idx, seq-idx) for each edge
      all_edges[idx] = numpy.vstack((all_edges[idx], numpy.ones((1, num_edges), dtype='uint32') * idx))

    start_end_states = numpy.empty((2, len(all_num_states)), dtype='uint32')
    state_offset = 0
    for idx, num_states in enumerate(all_num_states):
      start_end_states[0, idx] = state_offset
      start_end_states[1, idx] = state_offset + num_states - 1
      state_offset += num_states

    return numpy.hstack(all_edges), numpy.hstack(all_weights), start_end_states

  def get_free_instance(self):
    """
    :rtype: SprintSubprocessInstance|None
    """
    for inst in self.instances:
      if not inst.is_calculating:
        return inst
    return self._maybe_create_new_instance()




import _setup_test_env  # noqa
from nose.tools import assert_equal, assert_true
from returnn.engine.batch import Batch
from returnn.config import Config
import returnn.util.basic as util
from returnn.datasets.sprint import ExternSprintDataset
import os
import sys
import unittest
from returnn.util import better_exchook

dummyconfig_dict = {
  "num_inputs": 2,
  "num_outputs": 3,
  "hidden_size": (1,),
  "hidden_type": "forward",
  "activation": "relu",
  "bidirectional": False,
}


os.chdir((os.path.dirname(__file__) or ".") + "/..")
assert os.path.exists("rnn.py")
sprintExecPath = "tests/DummySprintExec.py"


def generate_batch(seq_idx, dataset):
  batch = Batch()
  batch.add_frames(seq_idx=seq_idx, seq_start_frame=0, length=dataset.get_seq_length(seq_idx))
  return batch


def test_read_all():
  config = Config()
  config.update(dummyconfig_dict)
  print("Create ExternSprintDataset")
  python2_exec = util.which("python2")
  if python2_exec is None:
    raise unittest.SkipTest("python2 not found")
  num_seqs = 4
  dataset = ExternSprintDataset(
    [python2_exec, sprintExecPath],
    "--*.feature-dimension=2 --*.trainer-output-dimension=3 "
    "--*.crnn-dataset=DummyDataset(2,3,num_seqs=%i,seq_len=10)" % num_seqs)
  dataset.init_seq_order(epoch=1)
  seq_idx = 0
  while dataset.is_less_than_num_seqs(seq_idx):
    dataset.load_seqs(seq_idx, seq_idx + 1)
    for key in dataset.get_data_keys():
      value = dataset.get_data(seq_idx, key)
      print("seq idx %i, data %r: %r" % (seq_idx, key, value))
    seq_idx += 1
  assert seq_idx == num_seqs


def test_assign_dev_data():
  config = Config()
  config.update(dummyconfig_dict)
  print("Create ExternSprintDataset")
  dataset = ExternSprintDataset(
    [sys.executable, sprintExecPath],
    "--*.feature-dimension=2 --*.trainer-output-dimension=3 --*.crnn-dataset=DummyDataset(2,3,num_seqs=4,seq_len=10)")
  dataset.init_seq_order(epoch=1)
  assert_true(dataset.is_less_than_num_seqs(0))
  recurrent = False
  batch_generator = dataset.generate_batches(recurrent_net=recurrent, batch_size=5)
  batches = batch_generator.peek_next_n(2)
  assert_equal(len(batches), 2)


def test_window():
  input_dim = 2
  output_dim = 3
  num_seqs = 2
  seq_len = 5
  window = 3
  dataset_kwargs = dict(
    sprintTrainerExecPath=[sys.executable, sprintExecPath],
    sprintConfigStr=" ".join([
      "--*.feature-dimension=%i" % input_dim,
      "--*.trainer-output-dimension=%i" % output_dim,
      "--*.crnn-dataset=DummyDataset(input_dim=%i,output_dim=%i,num_seqs=%i,seq_len=%i)" % (
        input_dim, output_dim, num_seqs, seq_len)]))
  dataset1 = ExternSprintDataset(**dataset_kwargs)
  dataset2 = ExternSprintDataset(window=window, **dataset_kwargs)
  try:
    dataset1.init_seq_order(epoch=1)
    dataset2.init_seq_order(epoch=1)
    dataset1.load_seqs(0, 1)
    dataset2.load_seqs(0, 1)
    assert_equal(dataset1.get_data_dim("data"), input_dim)
    assert_equal(dataset2.get_data_dim("data"), input_dim * window)
    data1 = dataset1.get_data(0, "data")
    data2 = dataset2.get_data(0, "data")
    assert_equal(data1.shape, (seq_len, input_dim))
    assert_equal(data2.shape, (seq_len, window * input_dim))
    data2a = data2.reshape(seq_len, window, input_dim)
    print("data1:")
    print(data1)
    print("data2:")
    print(data2)
    print("data1[0]:")
    print(data1[0])
    print("data2[0]:")
    print(data2[0])
    print("data2a[0,0]:")
    print(data2a[0, 0])
    assert_equal(list(data2a[0, 0]), [0] * input_dim)  # zero-padded left
    assert_equal(list(data2a[0, 1]), list(data1[0]))
    assert_equal(list(data2a[0, 2]), list(data1[1]))
    assert_equal(list(data2a[1, 0]), list(data1[0]))
    assert_equal(list(data2a[1, 1]), list(data1[1]))
    assert_equal(list(data2a[1, 2]), list(data1[2]))
    assert_equal(list(data2a[-1, 2]), [0] * input_dim)  # zero-padded right
  finally:
    dataset1._exit_handler()
    dataset2._exit_handler()


def test_py2_client():
  # like test_read_all
  config = Config()
  config.update(dummyconfig_dict)
  print("Create ExternSprintDataset")
  python2_exec = util.which("python2")
  if python2_exec is None:
    raise unittest.SkipTest("python2 not found")
  num_seqs = 4
  dataset = ExternSprintDataset(
    [python2_exec, sprintExecPath],
    "--*.feature-dimension=2 --*.trainer-output-dimension=3 "
    "--*.crnn-dataset=DummyDataset(2,3,num_seqs=%i,seq_len=10)" % num_seqs)
  dataset.init_seq_order(epoch=1)
  seq_idx = 0
  while dataset.is_less_than_num_seqs(seq_idx):
    dataset.load_seqs(seq_idx, seq_idx + 1)
    for key in dataset.get_data_keys():
      value = dataset.get_data(seq_idx, key)
      print("seq idx %i, data %r: %r" % (seq_idx, key, value))
    seq_idx += 1
  assert seq_idx == num_seqs


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




"""
Here are all subprocess, threading etc related utilities,
most of them quite low level.
"""

from __future__ import print_function
from threading import Lock
import sys

import os
import io
import pickle
import types
import struct
import marshal
from importlib import import_module
import numpy
try:
  from _multiprocessing import Connection
except ImportError:
  from multiprocessing.connection import Connection

PY3 = sys.version_info[0] >= 3

if PY3:
  from io import BytesIO
else:
  # noinspection PyUnresolvedReferences,PyCompatibility
  from StringIO import StringIO as BytesIO


_abs_mod_file = os.path.abspath(__file__)


class SharedMem:
  class ShmException(Exception): pass
  class CCallException(ShmException): pass

  if sys.platform != "win32":
    import ctypes
    import ctypes.util

    libc_so = ctypes.util.find_library('c')
    libc = ctypes.CDLL(libc_so, use_errno=True)
    shm_key_t = ctypes.c_int
    IPC_PRIVATE = 0
    IPC_RMID = 0

    # int shmget(key_t key, size_t size, int shmflg);
    shmget = libc.shmget
    shmget.restype = ctypes.c_int
    shmget.argtypes = (shm_key_t, ctypes.c_size_t, ctypes.c_int)
    # void* shmat(int shmid, const void *shmaddr, int shmflg);
    shmat = libc.shmat
    shmat.restype = ctypes.c_void_p
    shmat.argtypes = (ctypes.c_int, ctypes.c_void_p, ctypes.c_int)
    # int shmdt(const void *shmaddr);
    shmdt = libc.shmdt
    shmdt.restype = ctypes.c_int
    shmdt.argtypes = (ctypes.c_void_p,)
    # int shmctl(int shmid, int cmd, struct shmid_ds *buf);
    shmctl = libc.shmctl
    shmctl.restype = ctypes.c_int
    shmctl.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_void_p)
    # void* memcpy( void *dest, const void *src, size_t count );
    memcpy = libc.memcpy
    memcpy.restype = ctypes.c_void_p
    memcpy.argtypes = (ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)

    @classmethod
    def check_ccall_error(cls, check, f):
      import ctypes
      if not check:
        errno = ctypes.get_errno()
        errstr = os.strerror(errno)
        raise cls.CCallException("SharedMem: %s failed with error %i (%s)" % (f, errno, errstr))

    @classmethod
    def is_shmget_functioning(cls):
      shmid = cls.shmget(cls.IPC_PRIVATE, 4 * 1024 * 1024, 0o600)
      if shmid <= 0:
        return False
      cls.shmctl(shmid, cls.IPC_RMID, 0)
      return True

    def __init__(self, size, shmid=None):
      self.size = size
      self.shmid = None
      self.ptr = None
      if shmid is None:
        self.is_creator = True
        self.shmid = self.shmget(self.IPC_PRIVATE, self.size, 0o600)
        self.check_ccall_error(self.shmid > 0, "shmget")
        print("SharedMem[pid %i]: New shmid: %i (size %i)" % (os.getpid(), self.shmid, self.size))
        import atexit
        atexit.register(self.remove)
      else:
        self.is_creator = False
        self.shmid = shmid
        assert self.shmid > 0
      self.ptr = self.shmat(self.shmid, 0, 0)
      self.check_ccall_error(self.ptr != self.ctypes.c_void_p(-1).value, "shmat")
      self.check_ccall_error(self.ptr > 0, "shmat")

    def remove(self):
      if self.ptr:
        self.shmdt(self.ptr)
        self.ptr = None
      if self.shmid and self.shmid > 0:
        if self.is_creator:
          print("SharedMem[pid %i]: Removing shmid %i (size %i)" % (os.getpid(), self.shmid, self.size))
          self.shmctl(self.shmid, self.IPC_RMID, 0)
        self.shmid = None

    def __del__(self):
      self.remove()

    def __getstate__(self):
      return {"size": self.size, "shmid": self.shmid}

    def __setstate__(self, state):
      self.__init__(**state)

    def __repr__(self):
      return "<SharedMem shmid=%r size=%r is_creator=%r>" % (self.shmid, self.size, self.is_creator)


def next_power_of_two(n):
  return 2 ** (int(n - 1).bit_length())


class SharedNumpyArray:
  """
  This class provides a way to create Numpy arrays in shared memory.
  It adds some logic to mark whether some shared memory segment can be reused
  - that is when the client marks it as unused.

  Note that there are a few similar Python modules:
    https://pypi.python.org/pypi/SharedArray
    https://parad0x.org/git/python/shared-array/about
    https://bitbucket.org/cleemesser/numpy-sharedmem/src
    https://stackoverflow.com/questions/5033799/how-do-i-pass-large-numpy-arrays
    https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory
  """

  # cls members
  ServerLock = Lock()
  ServerInstances = set()
  ServerArrayId = 0
  class TooMuchInstances(SharedMem.ShmException): pass
  ExtraSpaceBytes = 4096
  # local members
  is_server = False
  mem = None
  shape, strides, typestr = None, None, None

  @staticmethod
  def numpy_strides_for_fortran(shape, typestr):
    itemsize = numpy.dtype(typestr).itemsize
    strides = [itemsize]
    for s in shape:
      strides += [strides[-1] * s]
    strides = strides[:-1]
    return tuple(strides)

  @staticmethod
  def numpy_strides_for_c_contiguous(shape, typestr):
    itemsize = numpy.dtype(typestr).itemsize
    strides = [numpy.prod(shape[i + 1:], dtype="int") * itemsize for i in range(len(shape))]
    return tuple(strides)

  @classmethod
  def needed_mem_size(cls, shape, typestr):
    itemsize = numpy.dtype(typestr).itemsize
    mem_size = cls.ExtraSpaceBytes + itemsize * numpy.prod(shape)
    return mem_size

  @classmethod
  def as_shared(cls, array):
    assert isinstance(array, numpy.ndarray)
    if isinstance(array.base, SharedNumpyArray):
      assert array.base.is_in_use()
      return array.base
    return cls.create_copy(array)

  @classmethod
  def create_copy(cls, array):
    assert isinstance(array, numpy.ndarray)
    array_intf = array.__array_interface__
    shape = array_intf["shape"]
    strides = array_intf["strides"]
    typestr = array_intf["typestr"]
    if array.flags.c_contiguous or array.flags.f_contiguous:
      pass  # ok, we can reuse it like that
    else:
      assert strides
      # Use some similar strides so that the copying might be faster.
      if strides[0] == array.itemsize:
        strides = cls.numpy_strides_for_fortran(shape=shape, typestr=typestr)
      else:
        strides = None  # C-contiguous
    inst = cls.create_new(shape=shape, strides=strides, typestr=typestr)
    inst.create_numpy_array()[...] = array
    assert inst._get_sanity_check_flag_ref().value == 42
    assert inst.is_in_use()
    return inst

  @classmethod
  def create_new(cls, shape, strides, typestr):
    needed_mem_size = cls.needed_mem_size(shape=shape, typestr=typestr)
    with cls.ServerLock:
      for inst in cls.ServerInstances:
        assert isinstance(inst, SharedNumpyArray)
        assert inst._get_sanity_check_flag_ref().value == 42
        if inst.is_in_use(): continue
        if inst.mem.size < needed_mem_size:
          inst._init_mem(shape=shape, typestr=typestr)
        # We can reuse it.
        inst._set_new_array_id()
        inst._set_is_used(1)
        inst._set_numpy_format(shape=shape, strides=strides, typestr=typestr)
        return inst
    return cls(shape=shape, strides=strides, typestr=typestr)

  @classmethod
  def _get_new_array_id(cls):
    array_id = cls.ServerArrayId
    cls.ServerArrayId += 1
    return array_id

  def _set_new_array_id(self):
    assert self.is_server
    self.array_id = self._get_new_array_id()
    self._get_array_id_ref().value = self.array_id

  def __init__(self, shape, strides, typestr, mem=None, array_id=None):
    if not mem:
      assert array_id is None
      if len(self.ServerInstances) >= SharedMemNumpyConfig["max_server_instances"]:
        raise self.TooMuchInstances("too much instances (%i)" % len(self.ServerInstances))
      self.is_server = True
      self._init_mem(shape=shape, typestr=typestr)
      self._set_new_array_id()
      self._set_is_used(1)
    else:
      assert array_id is not None
      self.is_server = False
      self.array_id = array_id
      mem_size = self.needed_mem_size(shape=shape, typestr=typestr)
      assert isinstance(mem, SharedMem)
      assert mem.size >= mem_size
      assert mem.shmid > 0
      assert mem.ptr > 0
      self.mem = mem
      assert self._get_sanity_check_flag_ref().value == 42
      assert self._get_array_id_ref().value == self.array_id
      assert self.is_in_use()
    self._set_numpy_format(shape=shape, strides=strides, typestr=typestr)
    if self.is_server:
      with self.ServerLock:
        self.ServerInstances.add(self)

  def _set_numpy_format(self, shape, strides, typestr):
    itemsize = numpy.dtype(typestr).itemsize
    if strides:
      assert all([st > 0 for st in strides])
      assert sum([st * (sh - 1) for (st, sh) in zip(strides, shape)]) + itemsize == numpy.prod(shape) * itemsize
    self.shape = shape
    self.strides = strides
    self.typestr = typestr

  def _init_mem(self, shape, typestr):
    assert self.is_server
    if self.mem:
      self.mem.remove()
      self.mem = None
    assert numpy.prod(shape) > 0
    mem_size = next_power_of_two(self.needed_mem_size(shape=shape, typestr=typestr))
    mem_size = max(SharedMemNumpyConfig["min_shared_mem_size"], mem_size)
    self.mem = SharedMem(size=mem_size)
    self._get_sanity_check_flag_ref().value = 42

  def get_numpy_array_data_ptr(self):
    assert self.mem.ptr > 0
    return self.mem.ptr + self.ExtraSpaceBytes

  @property
  def __array_interface__(self):
    assert self.shape
    # https://docs.scipy.org/doc/numpy/reference/arrays.interface.html
    return {
      "data": (self.get_numpy_array_data_ptr(), False),
      "shape": self.shape,
      "strides": self.strides,
      'typestr': self.typestr,
      "version": 3
    }

  def create_numpy_array(self):
    assert self._get_sanity_check_flag_ref().value == 42
    assert self._get_array_id_ref().value == self.array_id
    assert self.is_in_use()
    a = numpy.array(self, copy=False)
    assert a.__array_interface__["data"][0] == self.get_numpy_array_data_ptr()
    assert not a.flags.owndata, "a.__array_interface__ = %r" % a.__array_interface__
    assert a.base is self
    assert a.nbytes + self.ExtraSpaceBytes <= self.mem.size
    assert sum([st * (sh - 1) for (st, sh) in zip(a.strides, a.shape)]) + a.itemsize == numpy.prod(a.shape) * a.itemsize == a.nbytes
    return a

  def _get_sanity_check_flag_ref(self):
    assert self.mem.ptr > 0
    import ctypes
    return ctypes.cast(ctypes.c_void_p(self.mem.ptr), ctypes.POINTER(ctypes.c_uint64)).contents

  def _get_array_id_ref(self):
    assert self.mem.ptr > 0
    import ctypes
    return ctypes.cast(ctypes.c_void_p(self.mem.ptr + 8), ctypes.POINTER(ctypes.c_uint64)).contents

  def _get_in_use_flag_ref(self):
    assert self.mem.ptr > 0
    import ctypes
    return ctypes.cast(ctypes.c_void_p(self.mem.ptr + 16), ctypes.POINTER(ctypes.c_uint64)).contents

  def _set_is_used(self, n):
    self._get_in_use_flag_ref().value = n

  def is_in_use(self):
    return self._get_in_use_flag_ref().value > 0

  def set_unused(self):
    if self.is_server: return
    if self.mem:
      self._set_is_used(0)
      self.mem.remove()
      self.mem = None

  def __getstate__(self):
    return {
      "shape": self.shape, "strides": self.strides, "typestr": self.typestr,
      "mem": self.mem, "array_id": self.array_id
    }

  def __setstate__(self, state):
    self.__init__(**state)

  def __del__(self):
    # On the server side, we will get deleted at program end
    # because we are referenced in the global SharedNumpyArray.ServerInstances.
    # On the client side, we will get deleted once we are not used anymore.
    # Note that self.array holds a reference to self.
    self.set_unused()

  def __repr__(self):
    return "<%s is_server=%r state=%r>" % (self.__class__.__name__, self.is_server, self.__getstate__())


def attrChain(base, *attribs, **kwargs):
  default = kwargs.get("default", None)
  obj = base
  for attr in attribs:
    if obj is None: return default
    obj = getattr(obj, attr, None)
  if obj is None: return default
  return obj


# This is needed in some cases to avoid pickling problems with bounded funcs.
def funcCall(attrChainArgs, args=()):
  f = attrChain(*attrChainArgs)
  return f(*args)


Unpickler = pickle.Unpickler
if PY3:
  def get_func_closure(f): return f.__closure__
  # (code, globals[, name[, argdefs[, closure]]])
  def get_func_tuple(f):
    return (
      f.__code__,
      f.__globals__,
      f.__name__,
      f.__defaults__,
      f.__closure__,
    )
else:
  def get_func_closure(f): return f.func_closure
  def get_func_tuple(f):
    return (
      f.func_code,
      f.func_globals,
      f.func_name,
      f.func_defaults,
      f.func_closure,
    )
_closure = (lambda x: lambda: x)(0)
# noinspection PyUnresolvedReferences
_cell = get_func_closure(_closure)[0]
CellType = type(_cell)
ModuleType = type(sys)
# noinspection PyUnresolvedReferences
DictType = dict if PY3 else types.DictionaryType

if PY3:
  class BufferType: "Dummy"
  def make_buffer(*args): assert False
else:
  # noinspection PyUnresolvedReferences
  make_buffer = buffer
  # noinspection PyUnresolvedReferences
  BufferType = types.BufferType
  def bytes(x, *args): return str(x)

if PY3:
  _old_style_class = None
  class OldStyleClass: "Dummy"
  class _new_style_class: pass
  NewStyleClass = type
else:
  class _old_style_class: pass
  class _new_style_class(object): pass
  OldStyleClass = type(_old_style_class)  # == types.ClassType (classobj)
  NewStyleClass = type(_new_style_class)  # (type)

def makeFuncCell(value):
  return get_func_closure((lambda: value))[0]

def getModuleDict(modname, path=None):
  """
  :param str modname: such that "import <modname>" would work
  :param list[str] path: sys.path
  :return: the dict of the mod
  :rtype: dict[str]
  """
  try:
    mod = import_module(modname)
  except ImportError:
    # Try again with extended sys.path.
    assert path
    for p in path:
      if p not in sys.path:
        sys.path.append(p)
    mod = import_module(modname)
  return mod.__dict__

def getModNameForModDict(obj):
  """
  :type obj: dict
  :rtype: str | None
  :returns The module name or None. It will not return '__main__' in any case
  because that likely will not be the same in the unpickling environment.
  Also see: https://stackoverflow.com/questions/56171796/
  """
  if "__name__" not in obj:
    return None  # this is not a module
  mod_name = obj["__name__"]
  if mod_name == "__main__":
    return None
  if mod_name not in sys.modules:
    return None  # does not look like we have it loaded
  mod = sys.modules[mod_name]
  if mod.__dict__ is obj:
    return mod_name
  return None

def getNormalDict(d):
  """
  :type d: dict[str] | dictproxy
  :rtype: dict[str]
  It also removes getset_descriptor. New-style classes have those.
  """
  r = {}
  for k, v in d.items():
    if isinstance(v, types.GetSetDescriptorType): continue
    r[k] = v
  return r

def make_numpy_ndarray_fromstring(s, dtype, shape):
  return numpy.fromstring(s, dtype=dtype).reshape(shape)


SharedMemNumpyConfig = {
  "enabled": False,
  "auto_pickling_min_size": 8 * 1024 * 1024,  # 8MB
  "min_shared_mem_size": 32 * 1024 * 1024,  # 32MB
  "max_server_instances": 10,
}

def use_shared_mem_for_numpy_array(obj):
  assert isinstance(obj, numpy.ndarray)
  if obj.shape == ():  # scalar
    return False  # cannot use shared memory because it will always use its own memory
  if isinstance(obj.base, SharedNumpyArray):
    assert obj.base.is_in_use()
    return True
  if not SharedMemNumpyConfig["enabled"]:
    return False
  return obj.nbytes >= SharedMemNumpyConfig["auto_pickling_min_size"]

def numpy_set_unused(v):
  """
  :param numpy.ndarray v: array which will be marked as not-used-anymore
  This will tell mechanisms like SharedNumpyArray that it can reuse the memory.
  On the client side, this will even unmap the memory, so any further access
  to it will cause a SEGFAULT.
  """
  if v is None: return
  assert isinstance(v, numpy.ndarray)
  if isinstance(v.base, SharedNumpyArray):
    assert v.base.is_in_use()  # must not be called multiple times
    v.base.set_unused()

def numpy_copy_and_set_unused(v):
  """
  :param dict[str,numpy.ndarray|object] | numpy.ndarray | object v: object to be handled
  If v is a dict, we will return a new copied dict where every value is mapped through numpy_copy_and_set_unused.
  If v is a numpy.ndarray and its base is a SharedNumpyArray, we will copy it and
    call numpy_set_unused on the old value.
  If v is a numpy.ndarray and its base is not a SharedNumpyArray, we will just return it as it is and do nothing.
  In all other cases, we will also just return the object as it is and do nothing.
  """
  if isinstance(v, numpy.ndarray):
    if isinstance(v.base, SharedNumpyArray):
      newv = v.copy(order="A")
      numpy_set_unused(v)
      return newv
    return v
  if isinstance(v, dict):
    return {k: numpy_copy_and_set_unused(vv) for (k, vv) in v.items()}
  return v

def numpy_alloc(shape, dtype, fortran_for_shared=False):
  """
  If EnableAutoNumpySharedMemPickling is True, this will allocate a Numpy array
  in shared memory so we avoid a copy later on when this Numpy array would
  be transferred to another process via pickling.
  """
  if SharedMemNumpyConfig["enabled"]:
    dtype = numpy.dtype(dtype)
    typestr = dtype.str
    strides = None
    if fortran_for_shared:
      strides = SharedNumpyArray.numpy_strides_for_fortran(shape=shape, typestr=typestr)
    try:
      return SharedNumpyArray.create_new(shape=shape, strides=strides, typestr=typestr)
    except SharedMem.ShmException as e:
      print("numpy_alloc: SharedMem exception: %s" % e)
  # Fallback.
  return numpy.ndarray(shape, dtype=dtype)


try:
  _BasePickler = pickle._Pickler  # use the pure Python implementation
except AttributeError:
  _BasePickler = pickle.Pickler


class Pickler(_BasePickler):
  """
  We extend the standard Pickler to be able to pickle some more types,
  such as lambdas and functions, code, func cells, buffer and more.
  """

  def __init__(self, *args, **kwargs):
    if not "protocol" in kwargs:
      kwargs["protocol"] = pickle.HIGHEST_PROTOCOL
    _BasePickler.__init__(self, *args, **kwargs)
  dispatch = _BasePickler.dispatch.copy()

  def save_func(self, obj):
    try:
      self.save_global(obj)
      return
    except pickle.PicklingError:
      pass
    assert type(obj) is types.FunctionType
    self.save(types.FunctionType)
    self.save(get_func_tuple(obj))
    self.write(pickle.REDUCE)
    if id(obj) not in self.memo:  # Could be if we recursively landed here. See also pickle.save_tuple().
      self.memoize(obj)
  dispatch[types.FunctionType] = save_func

  def save_method(self, obj):
    try:
      self.save_global(obj)
      return
    except pickle.PicklingError:
      pass
    assert type(obj) is types.MethodType
    self.save(types.MethodType)
    if PY3:
      self.save((obj.__func__, obj.__self__))
    else:
      self.save((obj.im_func, obj.im_self, obj.im_class))
    self.write(pickle.REDUCE)
    self.memoize(obj)
  dispatch[types.MethodType] = save_method

  def save_code(self, obj):
    assert type(obj) is types.CodeType
    self.save(marshal.loads)
    self.save((marshal.dumps(obj),))
    self.write(pickle.REDUCE)
    self.memoize(obj)
  dispatch[types.CodeType] = save_code

  def save_cell(self, obj):
    assert type(obj) is CellType
    self.save(makeFuncCell)
    self.save((obj.cell_contents,))
    self.write(pickle.REDUCE)
    self.memoize(obj)
  dispatch[CellType] = save_cell

  # We also search for module dicts and reference them.
  # This is for FunctionType.func_globals.
  def intellisave_dict(self, obj):
    modname = getModNameForModDict(obj)
    if modname:
      self.save(getModuleDict)
      self.save((modname, sys.path))
      self.write(pickle.REDUCE)
      self.memoize(obj)
      return
    self.save_dict(obj)
  dispatch[DictType] = intellisave_dict

  def save_module(self, obj):
    modname = getModNameForModDict(obj.__dict__)
    if modname:
      self.save(import_module)
      self.save((modname,))
      self.write(pickle.REDUCE)
      self.memoize(obj)
      return
    # We could maybe construct it manually. For now, just fail.
    raise pickle.PicklingError('cannot pickle module %r' % obj)
  dispatch[ModuleType] = save_module

  def save_buffer(self, obj):
    self.save(buffer)
    self.save((str(obj),))
    self.write(pickle.REDUCE)
  dispatch[BufferType] = save_buffer

  def save_string(self, obj, pack=struct.pack):
    # Difference to base: We just always use BINSTRING (simpler)
    # and use a separate write for the obj itself.
    # For a huge obj, this avoids one unnecessary copy of the data.
    self.write(pickle.BINSTRING + pack("<i", len(obj)))
    self.write(bytes(obj, "utf8"))
  dispatch[str] = save_string

  def save_ndarray(self, obj):
    if use_shared_mem_for_numpy_array(obj):
      try:
        shared = SharedNumpyArray.as_shared(obj)
      except SharedMem.ShmException as e:
        print("SharedNumpyArray exception: %s" % e)
        # fallback to default
      else:
        self.save(shared.create_numpy_array)
        self.save(())
        self.write(pickle.REDUCE)
        return
    # For some reason, Numpy fromstring/tostring is faster than Numpy loads/dumps.
    self.save(make_numpy_ndarray_fromstring)
    self.save((obj.tostring(), str(obj.dtype), obj.shape))
    self.write(pickle.REDUCE)
  dispatch[numpy.ndarray] = save_ndarray

  def save_iobuffer_dummy(self, obj):
    # Not supported but we want to not fail and just store None.
    self.save_none(None)
  dispatch[io.BufferedReader] = save_iobuffer_dummy
  dispatch[io.BufferedWriter] = save_iobuffer_dummy

  # Overwrite to avoid the broken pickle.whichmodule() which might return "__main__".
  def save_global(self, obj, name=None):
    assert obj
    assert id(obj) not in self.memo
    if name is None:
      name = obj.__name__

    module = getattr(obj, "__module__", None)
    if module == "__main__" and globals().get(name, None) is obj:
      # Can happen if this is directly executed.
      module = __name__  # should be correct now
    if module is None or module == "__main__":
      module = pickle.whichmodule(obj, name)
    if module is None or module == "__main__":
      raise pickle.PicklingError(
          "Can't pickle %r: module not found: %s" % (obj, module))

    try:
      __import__(module)
      mod = sys.modules[module]
      klass = getattr(mod, name)
    except (ImportError, KeyError, AttributeError):
      raise pickle.PicklingError(
          "Can't pickle %r: it's not found as %s.%s" % (obj, module, name))
    else:
      if klass is not obj:
        raise pickle.PicklingError(
            "Can't pickle %r: it's not the same object as %s.%s" % (obj, module, name))

    assert "\n" not in module
    assert "\n" not in name
    self.write(pickle.GLOBAL + bytes(module + '\n' + name + '\n', "utf8"))
    self.memoize(obj)

  def save_type(self, obj):
    try:
      self.save_global(obj)
      return
    except pickle.PicklingError:
      pass
    # Some types in the types modules are not correctly referenced,
    # such as types.FunctionType. This is fixed here.
    for modname in ["types"]:
      moddict = sys.modules[modname].__dict__
      for modobjname,modobj in moddict.items():
        if modobj is obj:
          self.write(pickle.GLOBAL + bytes(modname + '\n' + modobjname + '\n', "utf8"))
          self.memoize(obj)
          return
    # Generic serialization of new-style classes.
    self.save(type)
    self.save((obj.__name__, obj.__bases__, getNormalDict(obj.__dict__)))
    self.write(pickle.REDUCE)
    self.memoize(obj)
  dispatch[NewStyleClass] = save_type

  # This is about old-style classes.
  def save_class(self, cls):
    try:
      # First try with a global reference. This works normally. This is the default original pickle behavior.
      self.save_global(cls)
      return
    except pickle.PicklingError:
      pass
    # It didn't worked. But we can still serialize it.
    # Note that this could potentially confuse the code if the class is reference-able in some other way
    # - then we will end up with two versions of the same class.
    self.save(types.ClassType)
    self.save((cls.__name__, cls.__bases__, cls.__dict__))
    self.write(pickle.REDUCE)
    self.memoize(cls)
    return
  dispatch[OldStyleClass] = save_class

  # avoid pickling instances of ourself. this mostly doesn't make sense and leads to trouble.
  # however, also doesn't break. it mostly makes sense to just ignore.
  def __getstate__(self): return None
  def __setstate__(self, state): pass



#!/usr/bin/env python

"""
Dumps the network topology as JSON on stdout.
"""

from __future__ import print_function

import sys
import argparse
import json
import typing

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
from returnn.pretrain import pretrain_from_config
from returnn.config import network_json_from_config


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_filename, command_line_options):
  """
  :param str config_filename:
  :param list[str] command_line_options:
  """
  rnn.init_better_exchook()
  rnn.init_config(config_filename, command_line_options)
  global config
  config = rnn.config
  config.set("log", [])
  rnn.init_log()
  print("RETURNN dump-dataset starting up.", file=log.v3)
  rnn.init_config_json_network()


def main(argv):
  """
  Main entry.
  """
  arg_parser = argparse.ArgumentParser(description='Dump network as JSON.')
  arg_parser.add_argument('returnn_config_file')
  arg_parser.add_argument('--epoch', default=1, type=int)
  arg_parser.add_argument('--out', default="/dev/stdout")
  args = arg_parser.parse_args(argv[1:])
  init(config_filename=args.returnn_config_file, command_line_options=[])

  pretrain = pretrain_from_config(config)
  if pretrain:
    json_data = pretrain.get_network_json_for_epoch(args.epoch)
  else:
    json_data = network_json_from_config(config)

  f = open(args.out, 'w')
  print(json.dumps(json_data, indent=2, sort_keys=True), file=f)
  f.close()

  rnn.finalize()


if __name__ == '__main__':
  main(sys.argv)



"""
This file contains dataset implementations to have an easy to use
interface for using RETURNN for regression.
Applications are for example speech enhancement or mask estimations
"""

from __future__ import print_function

__author__ = 'menne'

import os
import numpy as np
import h5py
from collections import deque
from .cached2 import CachedDataset2
from returnn.datasets.basic import DatasetSeq
from .bundle_file import BundleFile
from .normalization_data import NormalizationData
from returnn.log import log


class StereoDataset(CachedDataset2):
  """The purpose of this dataset is to be a base dataset for datasets which
  have an easy to use interface for using RETURNN as a regression tool
  """

  def __init__(self, partition_epoch=1, **kwargs):
    """constructor"""
    super(StereoDataset, self).__init__(**kwargs)
    self._seq_index_list = None
    self._partition_epoch = partition_epoch
    self._current_partition = 0
    self._seqs_per_epoch = None

  def initialize(self):
    self._seq_overhead = self._get_total_number_of_sequences() % self._partition_epoch
    super(StereoDataset, self).initialize()

  @property
  def num_seqs(self):
    """returns the number of sequences of the dataset

    :rtype: int
    """
    if self._num_seqs is not None:
      return self._num_seqs
    raise NotImplementedError

  def _get_total_number_of_sequences(self):
    raise NotImplementedError

  @property
  def seqs_per_epoch(self):
    if self._seqs_per_epoch is None:
      self._seqs_per_epoch = self._get_total_number_of_sequences() // self._partition_epoch
    return self._seqs_per_epoch

  def _collect_single_seq(self, seq_idx):
    """returns the sequence specified by the index seq_idx

    :type seq_idx: int
    :rtype: DatasetSeq | None
    :returns DatasetSeq or None if seq_idx >= num_seqs.
    """
    raise NotImplementedError

  def _get_partition_size(self, partition):
    partition_size = self.seqs_per_epoch
    if partition == self._partition_epoch-1:
      partition_size += self._seq_overhead
    return partition_size

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :type epoch: int|None
    :param epoch: epoch number
    :param list[str]|None seq_list:
    :param list[int]|None seq_order:
    :param seq_list: only None is currently supported
    Initialize lists:
      self.seq_index  # sorted seq idx
    """
    super(StereoDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
    if epoch is None:
        self._seq_index_list = range(self.num_seqs)
        return True

    self._current_partition = (epoch - 1) % self._partition_epoch
    partition_size = self._get_partition_size(self._current_partition)

    if seq_list is not None or seq_order is not None:
      raise NotImplementedError('init_seq_order of StereoDataset does not support a predefined seq_list yet.')
    else:
      seq_index = self.get_seq_order_for_epoch(epoch, partition_size, lambda s: self.get_seq_length(s).get('data', None))

    self._seq_index_list = seq_index
    if epoch is not None:
      print("Reinitialize dataset seq order for epoch %i." % epoch, file=log.v4)

    return True


class StereoHdfDataset(StereoDataset):
  """A stereo dataset which needs an hdf file as input. The hdf file
  is supposed to always have group 'inputs' and for the training data it
  also needs to contain the group 'outputs'. Each group is supposed to
  contain one dataset per sequence. The names of the datasets are supposed
  to be consecutive numbers starting at 0.

  The datasets are 2D numpy arrays, where dimension 0 is the time axis and
  dimension 1 is the feature axis. Therefore dimension 0 of the 'input'
  dataset and the respective 'output' dataset need to be the same.
  """

  def __init__(self, hdfFile, num_outputs=None, normalizationFile=None,
               flag_normalizeInputs=True, flag_normalizeTargets=True,
               **kwargs):
    """Constructor

    :type hdfFile: str
    :param hdfFile: path to the hdf file. if a bundle file is given (*.bundle)
                    all hdf files listed in the bundle file will be used for
                    the dataset.
                    :see: BundleFile.BundleFile
    :type num_outputs: int
    :param num_outputs: this needs to be set if the stereo data hdf file
                        only contains 'inputs' data (e.g. for the extraction
                        process). Only if no 'outputs' data exists in the hdf
                        file num_outputs is used.
    :type normalizationFile: str | None
    :param normalizationFile: path to a HDF file with normalization data.
                              The file is optional: if it is not provided then
                              no normalization is performed.
                              :see: NormalizationData.NormalizationData
    :type flag_normalizeInputs: bool
    :param flag_normalizeInputs: if True then inputs will be normalized
                                 provided that the normalization HDF file has
                                 necessary datasets (i.e. mean and variance)
    :type flag_normalizeTargets: bool
    :param flag_normalizeTargets: if True then targets will be normalized
                                 provided that the normalization HDF file has
                                 necessary datasets (i.e. mean and variance)
    """
    super(StereoHdfDataset, self).__init__(**kwargs)

    self._flag_normalizeInputs = flag_normalizeInputs
    self._flag_normalizeTargets = flag_normalizeTargets
    # properties of the object which will be set further
    self.num_inputs = None
    self.num_outputs = None
    self._filePaths = None
    self._fileHandlers = None
    self._seqMap = None
    self._normData = None

    if not os.path.isfile(hdfFile):
      raise IOError(hdfFile + ' does not exits')
    self._initHdfFileHandlers(hdfFile)

    # set number of sequences in the dataset
    self._num_seqs = self._calculateNumberOfSequences()

    if normalizationFile is not None:
      self._setNormalization(normalizationFile)

    self._setInputAndOutputDimensions(num_outputs)

  def _initHdfFileHandlers(self, hdfFile):
    """Initialize HDF file handlers

    :type hdfFile: str
    :param hdfFile: path to an HDF file with sequences or to a bundle file
                    which should contain one path to an HDF file per line
                    :see: BundleFile.BundleFile
    """
    self._filePaths = []
    self._fileHandlers = []
    if hdfFile.endswith('.bundle'):  # a bundle file containing a list of hdf files is given
      bundle = BundleFile(hdfFile)
      for hdfFilePath in bundle.datasetFilePaths:
        self._filePaths.append(hdfFilePath)
        self._fileHandlers.append(h5py.File(hdfFilePath, 'r'))
    else:  # only a single hdf file is given
      self._filePaths.append(hdfFile)
      self._fileHandlers.append(h5py.File(hdfFile, 'r'))

  def _calculateNumberOfSequences(self):
    return self.seqs_per_epoch

  def _get_total_number_of_sequences(self):
    """Calculate and return the number of sequences in the dataset.
    This method also initializes a sequences map which maps sequence
    indices into HDF file handlers.

    :rtype: int
    :return: the number of sequences in the dataset
    """
    # initialize a sequence map to map the sequence index
    # from an hdf file into the corresponding
    # hdfFile and hdf-dataset name,
    # but it could e.g. be used for shuffling sequences as well
    self._seqMap = {}
    seqCounter = 0
    for fhIdx, fh in enumerate(self._fileHandlers):
      for k in fh['inputs'].keys():
        self._seqMap[seqCounter] = (fhIdx, k)
        seqCounter += 1
    return seqCounter

  def _setNormalization(self, normalizationFile):
    """Set optional normalization (mean and variance).
    Mean and variance are set only if they are provided.

    :type normalizationFile: string
    :param normalizationFile: path to an HDF normalization file which contains
                              optional datasets "mean" and "variance".
                              :see: NormalizationData.NormalizationData
    """
    if not os.path.isfile(normalizationFile):
      raise IOError(normalizationFile + ' does not exist')
    self._normData = NormalizationData(normalizationFile)

  def _setInputAndOutputDimensions(self, num_outputs):
    """Set properties which correspond to input and output dimensions.

    :type num_outputs: int
    :param num_outputs: dimensionality of output features. used only if
                        the dataset does not have output features. Or if output
                        features are sparse
    """
    someSequence = self._collect_single_seq(0)
    self.num_inputs = someSequence.get_data('data').shape[1]
    if 'outputs' in self._fileHandlers[0]:
      if len(someSequence.get_data('classes').shape) == 1:
        outputFeatDim = 1
      else:
        outputFeatDim = someSequence.get_data('classes').shape[1]
      if outputFeatDim == 1 and num_outputs is not None:
        self.num_outputs = {
          'classes': (num_outputs, outputFeatDim)
        }
      else:
        self.num_outputs = {
          'classes': (outputFeatDim, 2)
        }
    else:
      # in this case no output data is in the hdf file and
      # therfore the output dimension needs to be given
      # as an argument through the config file
      if num_outputs is None:
        raise ValueError(
          'if no output data is contained in StereoDataset'
          ' the output dimension has to be specified by num_outputs'
        )
      self.num_outputs = {'classes': (num_outputs, 2)}

  def get_data_dim(self, key):
    """This is copied from CachedDataset2 but the assertion is
    removed (see CachedDataset2.py)

    :type key: str
    :rtype: int
    :return: number of classes, no matter if sparse or not
    """
    if key == 'data':
      return self.num_inputs
    if key in self.num_outputs:
      d = self.num_outputs[key][0]
      return d
    self._load_something()
    if len(self.added_data[0].get_data(key).shape) == 1:
      return super(CachedDataset2, self).get_data_dim(key)  # unknown
    assert len(self.added_data[0].get_data(key).shape) == 2
    return self.added_data[0].get_data(key).shape[1]

  def __del__(self):
    """Closes HDF file handlers.
    """
    for fh in self._fileHandlers:
      try:
        fh.close()
      except Exception:
        pass

  @property
  def num_seqs(self):
    """Returns the number of sequences of the dataset

    :rtype: int
    :return: the number of sequences of the dataset.
    """
    # has been set during initialization of dataset ...
    if self._num_seqs is not None:
      return self._num_seqs
    # ... but for some reason _num_seqs is not set at specific points in the
    # execution of rnn.py therefore the following is a saveguard to fall back on
    self._num_seqs = self._calculateNumberOfSequences()
    return self._num_seqs

  def _collect_single_seq(self, seq_idx):
    """Returns the sequence specified by the index seq_idx.
    Normalization is applied to the input features if mean and variance
    have been specified during dataset creating (see the constructor).

    :type seq_idx: int
    :rtype: DatasetSeq | None
    :returns: None if seq_idx >= num_seqs or the corresponding sequence.
    """
    if self._seq_index_list is None:
        self.init_seq_order()

    if seq_idx >= len(self._seq_index_list):
      return None

    # map the seq_idx to the shuffled sequence indices
    shuf_seq_idx = self._seq_index_list[seq_idx]
    partition_offset = int(np.sum([self._get_partition_size(i1) for i1 in range(self._current_partition)]))
    shuf_seq_idx += partition_offset

    seqMapping = self._seqMap[shuf_seq_idx]
    fileIdx = seqMapping[0]
    datasetName = seqMapping[1]
    fileHandler = self._fileHandlers[fileIdx]
    inputFeatures = fileHandler['inputs'][datasetName][...]
    targets = None
    if 'outputs' in fileHandler:
      targets = fileHandler['outputs'][datasetName][...]

    # optional normalization
    if self._normData is not None:
      assert isinstance(self._normData, NormalizationData)
      if self._flag_normalizeInputs:
        inputFeatures = StereoHdfDataset._normalizeVector(inputFeatures, self._normData.inputMean, self._normData.inputVariance)
      if self._flag_normalizeTargets:
        targets = StereoHdfDataset._normalizeVector(targets, self._normData.outputMean, self._normData.outputVariance)

    # enforce float32 to enable Theano optimizations
    inputFeatures = inputFeatures.astype(np.float32)
    if (targets is not None) and targets.shape[1] > 1:
      targets = targets.astype(np.float32)
    elif targets.shape[1] == 1:
      targets = np.reshape(targets.astype(np.int32), (targets.shape[0],))

    return DatasetSeq(seq_idx, inputFeatures, targets)

  @staticmethod
  def _normalizeVector(v, mean, variance):
    """Helper method.
    Applies optional normalization to the given vector.

    :type v: numpy.ndarray | None
    :param v: vector if available or None otherwise
    :type mean: numpy.ndarray | None
    :param mean: mean
    :type variance: numpy.ndarray | None
    :param variance: variance
    :rtype: numpy.ndarray | None
    :return: normalized vector or None if it was None
    """
    if v is None:
      return v
    if mean is not None:
      v -= mean
    if variance is not None:
      v /= np.sqrt(variance)
    return v


class DatasetWithTimeContext(StereoHdfDataset):
  """This dataset composes a context feature by stacking together time frames.
  """

  def __init__(self, hdfFile, tau=1, **kwargs):
    """Constructor

    :type hdfFile: string
    :param hdfFile: see the StereoHdfDataset
    :type tau: int
    :param tau: how many time frames should be on the left and on the right.
                E.g. if tau = 2 then the context feature will be created
                by stacking two neighboring time frames from left and
                two neighboring time frames from right:
                newInputFeature = [ x_{t-2}, x_{t-1}, x_t, x_{t+1}, x_{t+2} ].
                In general new feature will have shape
                (2 * tau + 1) * originalFeatureDimensionality
                Output features are not changed.
    :type kwargs: dictionary
    :param kwargs: the rest of the arguments passed to the StereoHdfDataset
    """
    if tau <= 0:
      raise ValueError('context parameter tau should be greater than zero')
    self._tau = tau
    super(DatasetWithTimeContext, self).__init__(hdfFile, **kwargs)

  def _collect_single_seq(self, seq_idx):
    """this method implements stacking the features

    :type seq_idx: int
    :param seq_idx: index of a sequence
    :rtype: DatasetSeq
    :return: DatasetSeq
    """
    if seq_idx >= self.num_seqs:
      return None
    originalSeq = super(DatasetWithTimeContext, self)._collect_single_seq(
      seq_idx
    )
    inputFeatures = originalSeq.get_data('data')
    frames, bins = inputFeatures.shape
    leftContext = deque()
    rightContext = deque()
    inFeatWithContext = []
    for i in range(self._tau):
      leftContext.append(np.zeros(bins))
      if i + 1 < frames:
        rightContext.append(inputFeatures[i + 1, ...])
      else:
        rightContext.append(np.zeros(bins))
    for t in range(frames):
      f = inputFeatures[t, ...]
      newFeature = np.concatenate([np.concatenate(leftContext, axis=0), f, np.concatenate(rightContext, axis=0)], axis=0)
      inFeatWithContext.append(newFeature)
      leftContext.popleft()
      leftContext.append(f)
      rightContext.popleft()
      if t + 1 + self._tau < frames:
        rightContext.append(inputFeatures[t + 1 + self._tau, ...])
      else:
        rightContext.append(np.zeros(bins))
    inputFeatures = np.array(inFeatWithContext)
    targets = None
    if 'classes' in originalSeq.get_data_keys():
      targets = originalSeq.get_data('classes')
    return DatasetSeq(seq_idx, inputFeatures, targets)
